from __future__ import annotations

from chgnet.utils.common_utils import (
    AverageMeter,
    cuda_devices_sorted_by_free_mem,
    determine_device,
    mae,
    mkdir,
    read_json,
    write_json,
)
from chgnet.utils.vasp_utils import parse_vasp_dir, solve_charge_by_mag
